program mag_distortion_estimate
    use Globals


    implicit none

    call main()

    contains

    ! The actual subroutine
    subroutine main()

        use UserSuppliedParameters
        use UserInputs
        use ImageFiles
        use Images
        use ProgressBars
        use UsefulFunctions
        use Units
        use StringManipulations, only : IntegerToString

        ! Variables associated with user input

        type(UserInput)            ::   my_user_input
        type(UserSuppliedFilename) ::   input_filename
        type(UserSuppliedFilename) ::   output_amp_filename
        type(UserSuppliedFilename) ::   output_rot_amp_filename
        type(UserSuppliedFilename) ::   output_corrected_amp_filename

        type(UserSuppliedReal)     ::   pixel_size
        type(UserSuppliedReal)     ::   user_supplied_max_scale_factor
        type(UserSuppliedReal)     ::   user_supplied_scale_factor_step
        type(UserSuppliedReal)     ::   user_supplied_low_resolution
        type(UserSuppliedReal)     ::   user_supplied_high_resolution
        type(UserSuppliedReal)     ::   user_supplied_low_filter
        type(UserSuppliedReal)     ::   user_supplied_high_filter
        type(UserSuppliedReal)     ::   user_supplied_min_angle
        type(UserSuppliedReal)     ::   user_supplied_max_angle
        type(UserSuppliedReal)     ::   user_supplied_angle_step
        type(UserSuppliedInteger)  ::   user_supplied_amplitude_box_size

        type(UserSuppliedLogical)  ::   set_expert_options

        ! Other
        type(ImageFile)            ::  input_imagefile

        type(Image), allocatable   ::  image_stack(:)
        type(Image)                ::  sum_image
        type(Image)                ::  rot_average_image

        type(ProgressBar)          ::  my_progress_bar

        integer                    ::  number_of_input_images
        integer                    ::  image_counter
        integer                    ::  iterations_performed

        integer                    ::  square_size
        integer                    ::  cropped_size

        integer                    ::  inner_radius_gold_ring
        integer                    ::  outer_radius_gold_ring

        integer                    ::  stack_dimensions(3)

        integer                    ::  number_of_ticks

        logical                    ::  images_are_square

        integer                    ::  current_parameter


        integer                    ::  amplitude_box_size
        real                       ::  max_scale_factor
        real                       ::  high_pass_filter
        real                       ::  low_pass_filter
        real                       ::  inner_radius_resolution
        real                       ::  outer_radius_resolution

        real                       ::  brute_force_min_angle
        real                       ::  brute_force_max_angle
        real                       ::  brute_force_angle_step
        real                       ::  brute_force_scale_step

        real                       ::  image_size_scale_factor

        real                       ::  scaled_inner_radius
        real                       ::  scaled_outer_radius
        real                       ::  inner_radius_gold_ring_squared
        real                       ::  outer_radius_gold_ring_squared

        real                       ::  current_angle
        real                       ::  current_scale_factor

        real                       ::  best_correlation
        real                       ::  best_scale
        real                       ::  best_angle
        real                       ::  current_correlation

        real, allocatable          ::  angles(:)
        real, allocatable          ::  scale_factor(:)

        ! Initialise this_program with the program name and version number
        call this_program%Init('mag_distortion_estimate', '0.0.0','2015')


        ! Do the user input
        call my_user_input%Init(this_program%program_name)

        input_filename                  =   my_user_input%GetFilenameFromUser('Input stack filename', &
                                       'The input file, containing ~20 images of different areas of polycrystalline gold', &
                                       'input_filename', 'my_gold_images.mrc', .true.)

        output_amp_filename             =   my_user_input%GetFilenameFromUser('Output spectra filename', &
                                       'The output file containing the calculated ring image', &
                                       'output_amp_filename', 'my_amp.mrc', .false.)


        output_rot_amp_filename         =   my_user_input%GetFilenameFromUser('Output rot. avg. filename', &
                                       'The output file containing a rotationally averaged version of the ring image', &
                                       'output_rot_amp_filename', 'my_amp_rot.mrc', .false.)

        output_corrected_amp_filename   =   my_user_input%GetFilenameFromUser('Output corrected spectra', &
                                       'The output file containing a corrected_spectra', &
                                       'output_corrected_amp_filename', 'my_amp_corrected.mrc', .false.)


        pixel_size                      =   my_user_input%GetRealFromUser('Image Pixel size (Å)', &
                                        'The pixel size of the gold images in angstroms', &
                                        'pixel_size', '1.0')

        set_expert_options                  =  my_user_input%GetLogicalFromUser('Set expert options?', &
                                            'Allows tweaking parameters - should be unnecessary', &
                                            'set_expert_options', 'NO')

        if (set_expert_options%value) then

            user_supplied_low_resolution  =   my_user_input%GetRealFromUser('Low resolution limit (Å)', &
                                            'Low resolution limit for the search', &
                                            'low_resolution_limit', '2.5')

            user_supplied_high_resolution  =   my_user_input%GetRealFromUser('High resolution limit (Å)', &
                                            'High resolution limit for the search', &
                                            'high_resolution_limit', '2.1')


            user_supplied_max_scale_factor  =   my_user_input%GetRealFromUser('Max scale factor', &
                                            'maximum allowed scale factor', &
                                            'max_scale_factor', '0.03')

            user_supplied_scale_factor_step =   my_user_input%GetRealFromUser('Scale factor step size', &
                                            'Step size for the scale search', &
                                            'scale_factor_step', '0.0005')


            user_supplied_min_angle  =   my_user_input%GetRealFromUser('Minimum angle (degrees)', &
                                            'Minimum allowed angle', &
                                            'min_allowed_angle', '0.0')

            user_supplied_max_angle  =   my_user_input%GetRealFromUser('Maximum angle (degrees)', &
                                            'Maximum allowed angle', &
                                            'max_allowed_angle', '180.0')

            user_supplied_angle_step  =   my_user_input%GetRealFromUser('Angle step size (degrees)', &
                                            'Step size for the angle search', &
                                            'angle_step_size', '0.1')

            user_supplied_low_filter  =   my_user_input%GetRealFromUser('Low pass filter radius', &
                                            'Low pass filter radius for the amplitude bandpass filter', &
                                            'low_pass_filter', '0.2')


            user_supplied_high_filter  =   my_user_input%GetRealFromUser('High pass filter radius', &
                                            'High pass filter radius for the amplitude bandpass filter', &
                                            'high_pass_filter', '0.01')

            user_supplied_amplitude_box_size  =   my_user_input%GetIntegerFromUser('Amplitude box size', &
                                            'Box size for the calculated amplitudes', &
                                            'amplitude_box_size', '512')



            inner_radius_resolution = user_supplied_low_resolution%value
            outer_radius_resolution = user_supplied_high_resolution%value
            max_scale_factor = user_supplied_max_scale_factor%value
            brute_force_scale_step = user_supplied_scale_factor_step%value
            brute_force_min_angle = user_supplied_min_angle%value
            brute_force_max_angle = user_supplied_max_angle%value
            brute_force_angle_step = user_supplied_angle_step%value
            high_pass_filter = user_supplied_high_filter%value
            low_pass_filter = user_supplied_low_filter%value
            amplitude_box_size = user_supplied_amplitude_box_size%value

        else

            inner_radius_resolution = 2.5e0
            outer_radius_resolution = 2.1e0
            max_scale_factor = 0.03
            brute_force_scale_step = 0.0005e0
            brute_force_min_angle = 0.0e0
            brute_force_max_angle = 180.0e0
            brute_force_angle_step = 0.1e0

            amplitude_box_size = 512
            high_pass_filter = 0.01e0
            low_pass_filter = 0.2e0

        endif




        call my_user_input%UpdateDefaults()
        write(*,*)

        ! Open the input file
        call input_imagefile%Init(input_filename%value)

        ! Get dimensions of input
        number_of_input_images = input_imagefile%GetStackSize()
        stack_dimensions = input_imagefile%GetDimensions();

        ! crop size to make a square image..

        if (stack_dimensions(1) .eq. stack_dimensions(2)) then
            images_are_square = .true.
            square_size = stack_dimensions(1)
        elseif (stack_dimensions(1) .gt. stack_dimensions(2)) then
            images_are_square = .false.
            square_size = stack_dimensions(1)
        else
            images_are_square = .false.
            square_size = stack_dimensions(2)
        endif

        ! radius calculations

        outer_radius_gold_ring = (square_size / 2) * (pixel_size%value * 2.0) / outer_radius_resolution
        inner_radius_gold_ring = (square_size / 2) * (pixel_size%value * 2.0) / inner_radius_resolution

        ! what size do we need to crop it to before scaling it (we want outer resolution + ~30 pixels) to be amplitude image size

        image_size_scale_factor = real(amplitude_box_size) / real(outer_radius_gold_ring * 2) ! this is without any padding
        cropped_size = (outer_radius_gold_ring * 2) + int(30 / image_size_scale_factor) ! this should add on ~30 pixels

        ! don't pad..

        if (cropped_size .gt. square_size) cropped_size = square_size

        image_size_scale_factor = real(amplitude_box_size) / real(cropped_size) ! our final scale factor

        !write(*,*) 'square size = ', square_size, ', cropped size = ', cropped_size
        !write(*,*) 'Outer_radius_gold = ', outer_radius_gold_ring, ' , Inner = ', inner_radius_gold_ring
        !write(*,*) 'Image scale factor = ', image_size_scale_factor

        !scaled_inner_radius = real(amplitude_box_size / 2.0e0) * (pixel_size%value * 2.0e0) / inner_radius_resolution
        !scaled_outer_radius = real(amplitude_box_size / 2.0e0) * (pixel_size%value * 2.0e0) / outer_radius_resolution

        scaled_inner_radius = real(inner_radius_gold_ring) * image_size_scale_factor
        scaled_outer_radius = real(outer_radius_gold_ring) * image_size_scale_factor

        inner_radius_gold_ring_squared = scaled_inner_radius**2
        outer_radius_gold_ring_squared = scaled_outer_radius**2



        ! Allocation

        allocate(image_stack(number_of_input_images))


        write(*,'(a)') 'Reading images...'
        call my_progress_bar%Begin(number_of_input_images)

        do image_counter=1, number_of_input_images

            call image_stack(image_counter)%ReadFromDisk(input_imagefile, image_counter)
            call my_progress_bar%Update(image_counter)

        enddo

        call my_progress_bar%Finish()

        ! FFT

        iterations_performed = 0

        write(*,'(a)') 'Calculating Spectra...'
        call my_progress_bar%Begin(number_of_input_images)

        !$omp parallel default(shared) private(image_counter)
        !$omp do

        do image_counter = 1, number_of_input_images

            call image_stack(image_counter)%ZeroFloatAndNormalise(10.0)

            if (images_are_square .eq. .false.) then
                call image_stack(image_counter)%Resize([square_size, square_size, 1])
            endif


            call image_stack(image_counter)%ForwardFFT()
            image_stack(image_counter) = image_stack(image_counter)%GetAmplitudeSpectrum()


            ! crop the image to 2 x the outer radius

            call image_stack(image_counter)%Resize([cropped_size, cropped_size, 1])

            ! scale to 512x512 and filter..

            call image_stack(image_counter)%ForwardFFT
            call image_stack(image_counter)%Resize([amplitude_box_size, amplitude_box_size, 1])
            call image_stack(image_counter)%ApplyBandPassFilter(high_pass_filter, low_pass_filter)
            call image_stack(image_counter)%BackwardFFT


            !$omp atomic
            iterations_performed = iterations_performed + 1
            call my_progress_bar%Update(iterations_performed)

        enddo

        !$omp enddo
        !$omp end parallel

        ! Sum the spectra

        call sum_image%Allocate(dims=[amplitude_box_size, amplitude_box_size, 1],in_real_space=.true.)

        sum_image = 0.0e0

        do image_counter = 1, number_of_input_images

            call sum_image%AddImage(image_stack(image_counter))

        enddo



        call rot_average_image%Allocate(dims=[amplitude_box_size, amplitude_box_size, 1], in_real_space=.true.)
        rot_average_image = sum_image
        call rot_average_image%ApplyRotationalAverage

        call sum_image%ZeroFloatAndNormaliseWithinRing(60.0e0, real(amplitude_box_size * 2), 1.0e0)
        call sum_image%ZeroFloatAndNormaliseWithinRing(scaled_inner_radius, scaled_outer_radius, 1.0e0)
        call sum_image%WriteToDiskGivenFilename(output_amp_filename%value, 1)

        call rot_average_image%ZeroFloatAndNormaliseWithinRing(60.0e0, real(amplitude_box_size * 2), 1.0e0)
        call rot_average_image%ZeroFloatAndNormaliseWithinRing(scaled_inner_radius, scaled_outer_radius, 1.0e0)
        call rot_average_image%WriteToDiskGivenFilename(output_rot_amp_filename%value)


        call my_progress_bar%Finish()

        ! Now we do the search..
        !first a brute search..

        number_of_ticks = 0

        do current_angle = brute_force_min_angle, brute_force_max_angle, brute_force_angle_step
            do current_scale_factor = 0.0e0, max_scale_factor, brute_force_scale_step

                    number_of_ticks = number_of_ticks + 1

            enddo
        enddo

        ! fill in variables..


        if (allocated(image_stack)) deallocate(image_stack)

        allocate(angles(number_of_ticks))
        allocate(scale_factor(number_of_ticks))

        number_of_ticks = 0

        do current_angle = brute_force_min_angle, brute_force_max_angle, brute_force_angle_step
            do current_scale_factor = 0.0e0, max_scale_factor, brute_force_scale_step

                number_of_ticks = number_of_ticks + 1
                angles(number_of_ticks) = current_angle
                scale_factor(number_of_ticks) = current_scale_factor


            enddo
        enddo

        write(*,'(a)') 'Finding distortion parameters...'

        call my_progress_bar%Begin(number_of_ticks)
        iterations_performed = 0

        best_correlation = 0


        !$omp parallel default(shared) private(current_parameter, current_angle, current_scale_factor, current_correlation)
        !$omp do

        do current_parameter = 1, number_of_ticks

            current_angle = angles(current_parameter)
            current_scale_factor = scale_factor(current_parameter)

            current_correlation = calculate_correlation(sum_image, rot_average_image, current_angle, current_scale_factor, scaled_inner_radius, scaled_outer_radius)

            !$omp critical

            iterations_performed = iterations_performed + 1
            call my_progress_bar%Update(iterations_performed)

            if (current_correlation .gt. best_correlation) then

               best_correlation = current_correlation
               best_scale = current_scale_factor
               best_angle = current_angle

            endif

            !$omp endcritical

    enddo

    !$omp enddo

    !$omp end parallel

    call my_progress_bar%Finish


    call sum_image%CorrectMagnificationDistortion(rot_average_image, best_angle,  1.0e0 / (1.0e0 + (best_scale / 2.0e0)),  1.0e0 / (1.0e0 / (1.0e0 + (best_scale / 2.0e0))))
    call rot_average_image%ZeroFloatAndNormaliseWithinRing(60.0e0, real(amplitude_box_size * 2.0e0), 1.0e0)

    ! Write out the specta..

    call rot_average_image%WriteToDiskGivenFilename(output_corrected_amp_filename%value)


    write (*,*) 'The following distortion parameters were found :-'
    write (*,*)
    write (*,'(a,f0.1)') 'Distortion Angle     = ', best_angle
    write (*,'(a,f5.3)') 'Major Scale          = ', 1.0e0 + (best_scale / 2.0e0)
    write (*,'(a,f5.3)') 'Minor Scale          = ', 1.0e0 / (1.0e0 + (best_scale / 2.0e0))
    write (*,*)
    write (*,*) 'Stretch only parameters would be as follows :-'
    write (*,*)
    write (*,'(a,f0.1)') 'Distortion Angle     = ', best_angle
    write (*,'(a,f5.3)') 'Major Scale          = ', (1.0e0 + (best_scale / 2.0e0)) / (1.0e0 / (1.0e0 + (best_scale / 2.0e0)))
    write (*,'(a,f5.3)') 'Minor Scale          = ', 1.0e0
    write (*,'(a,f5.3)') 'Corrected Pixel Size = ', pixel_size%value / ((1.0e0 + (best_scale / 2.0e0)))
    write (*,*)
    write (*,'(a,f0.2, a)') 'The Total Distortion = ', ((1.0e0 + (best_scale / 2.0e0)) / (1.0e0 / (1.0e0 + (best_scale / 2.0e0))) - 1.0e0) * 100.0e0, '%'



    ! Deallocation
    call input_imagefile%Close()


    if (allocated(image_stack)) deallocate(image_stack)
    if (allocated(angles)) deallocate(angles)
    if (allocated(scale_factor)) deallocate(scale_factor)
    !if (allocated(minor_axis)) deallocate(minor_axis)

    ! All done
    call this_program%Terminate()

    end subroutine main


real function calculate_correlation(input_image, ref_image, angle, scale_factor, inner_radius, outer_radius)

    use Images
    use UsefulFunctions

    implicit none

    class(Image),   intent(inout)      ::  input_image
    class(Image),   intent(inout)      ::  ref_image
    real,           intent(in)         ::  scale_factor
    real,           intent(in)         ::  angle
    real,           intent(in)         ::  inner_radius
    real,           intent(in)         ::  outer_radius

    type(Image)                        ::  buffer_image
    type(Image)                        ::  buffer_image_rot

    integer                            ::  i,j
    integer                            ::  x,y
    integer                            ::  origin(3)

    real                               ::  major_axis_scale
    real                               ::  minor_axis_scale
    real                               ::  angle_in_radians
    real                               ::  inner_radius_squared
    real                               ::  outer_radius_squared

    real                               ::  major_scale_factor
    real                               ::  minor_scale_factor
    real                               ::  major_adjustment
    real                               ::  minor_adjustment
    real                               ::  current_square_radius
    real                               ::  aa_buffer
    real                               ::  ab_buffer
    real                               ::  bb_buffer

    origin = input_image%GetPhysicalAddressOfBoxCenter()

    major_axis_scale = 1.0e0 + (scale_factor / 2.0e0)
    minor_axis_scale = 1.0e0 / (1.0e0 + (scale_factor / 2.0e0))


    major_scale_factor = 1.0e0 / major_axis_scale
    minor_scale_factor = 1.0e0 / minor_axis_scale

    major_adjustment = real(input_image%GetLogicalDimension(1)) - (real(input_image%GetLogicalDimension(1))  * major_scale_factor)
    minor_adjustment = real(input_image%GetLogicalDimension(1)) - (real(input_image%GetLogicalDimension(1))  * minor_scale_factor)


    angle_in_radians = DegreesToRadians(angle)
    inner_radius_squared = inner_radius**2
    outer_radius_squared = outer_radius**2

    call buffer_image%Allocate(input_image)
    call buffer_image_rot%Allocate(input_image)

    ! Do the stretching

    call input_image%CorrectMagnificationDistortion(buffer_image, angle, major_scale_factor, minor_scale_factor)
    !call apply_correction(input_image, buffer_image, angle, scale_factor, inner_radius, outer_radius)

    ! now do the correlation

    buffer_image_rot = buffer_image
    call buffer_image_rot%ApplyRotationalAverage

    call buffer_image%ZeroFloatAndNormaliseWithinRing(inner_radius, outer_radius, 1.0e0)
    call buffer_image_rot%ZeroFloatAndNormaliseWithinRing(inner_radius, outer_radius, 1.0e0)

    ab_buffer = 0
    aa_buffer = 0
    bb_buffer = 0

    do j=1, buffer_image%GetLogicalDimension(2)
        y = (j-origin(2))**2
        do i=1,buffer_image%GetLogicalDimension(1)
            x = (i-origin(1))**2

            current_square_radius = real(x+y)

            if (current_square_radius .lt. outer_radius_squared .and. current_square_radius .gt. inner_radius_squared) then

                ab_buffer = ab_buffer + buffer_image%real_values(i, j, 1) * buffer_image_rot%real_values(i, j, 1)
                aa_buffer = aa_buffer + buffer_image%real_values(i, j, 1)**2
                bb_buffer = bb_buffer + buffer_image_rot%real_values(i, j, 1)**2

            endif
        enddo
    enddo

    calculate_correlation = ab_buffer / sqrt(aa_buffer * bb_buffer)

end function calculate_correlation

!
!subroutine apply_correction(input_image, output_image, angle, scale_factor, inner_radius, outer_radius)
!
!    use Images
!    use UsefulFunctions
!
!    implicit none
!
!    class(Image),   intent(inout)      ::  input_image
!    class(Image),   intent(inout)      ::  output_image
!    real,           intent(in)         ::  scale_factor
!    real,           intent(in)         ::  angle
!    real,           intent(in)         ::  inner_radius
!    real,           intent(in)         ::  outer_radius
!
!    integer                            ::  i,j
!    integer                            ::  origin(3)
!    integer                            ::  x,y
!
!    real                               ::  major_axis_scale
!    real                               ::  minor_axis_scale
!    real                               ::  angle_in_radians
!    real                               ::  inner_radius_squared
!    real                               ::  outer_radius_squared
!
!    real                               ::  major_scale_factor
!    real                               ::  minor_scale_factor
!    real                               ::  major_adjustment
!    real                               ::  minor_adjustment
!    real                               ::  current_square_radius
!    real                               ::  new_x
!    real                               ::  new_y
!    real                               ::  final_x
!    real                               ::  final_y
!    real                               ::  interpolated_value
!
!    major_axis_scale = 1.0e0 + (scale_factor / 2.0e0)
!    minor_axis_scale = 1.0e0 / (1.0e0 + (scale_factor / 2.0e0))
!
!    origin = input_image%GetPhysicalAddressOfBoxCenter()
!
!    major_scale_factor = 1.0e0 / major_axis_scale
!    minor_scale_factor = 1.0e0 / minor_axis_scale
!
!    major_adjustment = real(input_image%GetLogicalDimension(1)) - (real(input_image%GetLogicalDimension(1))  * major_scale_factor)
!    minor_adjustment = real(input_image%GetLogicalDimension(1)) - (real(input_image%GetLogicalDimension(1))  * minor_scale_factor)
!
!
!    angle_in_radians = DegreesToRadians(angle)
!    inner_radius_squared = inner_radius**2
!    outer_radius_squared = outer_radius**2
!
!
!    output_image%real_values = 0.0e0
!
!    do j=1, input_image%GetLogicalDimension(2)
!        y = (j - input_image%GetPhysicalIndexOfBoxCenter(2))
!        do i=1, input_image%GetLogicalDimension(1)
!
!            x = (i-input_image%GetPhysicalIndexOfBoxCenter(1))
!
!            current_square_radius = real(x**2+y**2)
!
!            if (current_square_radius .lt. outer_radius_squared .and. current_square_radius .gt. inner_radius_squared) then
!
!                !first off we do a rotation..
!
!                new_x =  real(j - input_image%GetPhysicalIndexOfBoxCenter(2))*sin(-angle_in_radians) &
!                        +real(i - input_image%GetPhysicalIndexOfBoxCenter(1))*cos(-angle_in_radians)
!                new_y =  real(j - input_image%GetPhysicalIndexOfBoxCenter(2))*cos(-angle_in_radians) &
!                        -real(i - input_image%GetPhysicalIndexOfBoxCenter(1))*sin(-angle_in_radians)
!
!            ! Scaling
!
!            !new_x = new_x * major_scale_factor
!            !new_y = new_y * minor_scale_factor
!
!            new_x = new_x + input_image%GetPhysicalIndexOfBoxCenter(1)
!            new_y = new_y + input_image%GetPhysicalIndexOfBoxCenter(2)
!
!
!            ! Rotate back
!
!            final_x = real(new_y - input_image%GetPhysicalIndexOfBoxCenter(2))*sin(angle_in_radians) &
!                     +real(new_x - input_image%GetPhysicalIndexOfBoxCenter(1))*cos(angle_in_radians)
!            final_y = real(new_y - input_image%GetPhysicalIndexOfBoxCenter(2))*cos(angle_in_radians) &
!                     -real(new_x - input_image%GetPhysicalIndexOfBoxCenter(1))*sin(angle_in_radians)
!
!            final_x = final_x + input_image%GetPhysicalIndexOfBoxCenter(1)
!            final_y = final_y + input_image%GetPhysicalIndexOfBoxCenter(2)
!
!            ! now the actual interpolation..
!            if (final_x .lt. 1 .or. final_x .gt. input_image%GetLogicalDimension(1)) then
!                output_image%real_values(i, j, 1) = 0
!            elseif (final_y .lt. 1 .or. final_y .gt. input_image%GetLogicalDimension(2)) then
!                output_image%real_values(i, j, 1) = 0
!            else
!                call input_image%GetRealValueByLinearInterpolationNoBoundsCheckImage(interpolated_value, final_x, final_y)
!                output_image%real_values(i, j, 1) = interpolated_value
!            endif
!
!            endif
!        enddo
!    enddo
!
!    call input_image%CorrectMagnificationDistortion(output_image, angle, major_scale_factor, minor_scale_factor, inner_radius, outer_radius)
!
!end subroutine apply_correction



end program mag_distortion_estimate

